'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.PortIsOccupiedError = exports.UnexpectedServerTerminationError = exports.ServerSpawnFailureError = undefined;
exports.sleep = sleep;
exports.default = spawnServer;

var _child_process = require('child_process');

var _portPid = require('./helpers/portPid');

var _portPid2 = _interopRequireDefault(_portPid);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

function sleep(timeout) {
  return new Promise(resolve => setTimeout(resolve, timeout));
}

class ServerSpawnFailureError extends Error {

  constructor(code, stdout, stderr) {
    super(`Server exited with the error code ${code}`);

    this.stderr = stderr.join('');
    this.stdout = stdout.join('');
  }
}

exports.ServerSpawnFailureError = ServerSpawnFailureError;
class UnexpectedServerTerminationError extends Error {

  constructor(stdout, stderr) {
    super('Unexpected server termination, probably you forgot to call listen()?');

    this.stderr = stderr.join('');
    this.stdout = stdout.join('');
  }
}

exports.UnexpectedServerTerminationError = UnexpectedServerTerminationError;
class PortIsOccupiedError extends Error {
  constructor(pid, port) {
    super(`Process with PID ${pid} is already listening on port ${port}`);
  }
}

exports.PortIsOccupiedError = PortIsOccupiedError;
function spawnServer(filepath, cwd, port = 3000, retryLimit = 10) {
  return new Promise((() => {
    var _ref = _asyncToGenerator(function* (resolve, reject) {
      try {
        let onExit = (() => {
          var _ref2 = _asyncToGenerator(function* (code) {
            // detect the cause, if there is another process listening on given port
            const pids = yield (0, _portPid2.default)(port);

            if (!pids.all.has(proc.pid) && pids.all.size > 0) {
              return finish(new PortIsOccupiedError([...pids.all][0], port));
            }

            if (code !== 0) {
              return finish(new ServerSpawnFailureError(code, stdout, stderr));
            }

            return finish(new UnexpectedServerTerminationError(stdout, stderr));
          });

          return function onExit(_x3) {
            return _ref2.apply(this, arguments);
          };
        })();

        const proc = (0, _child_process.fork)(filepath, {
          cwd,
          stdio: [null, null, null, 'ipc']
        });
        let finished = false;
        let stderr = [];
        let stdout = [];

        proc.stderr.on('data', function (buffer) {
          return stderr.push(buffer.toString('utf8'));
        });
        proc.stdout.on('data', function (buffer) {
          return stdout.push(buffer.toString('utf8'));
        });

        function onError(e) {
          finish(e);
        }

        function finish(error) {
          proc.removeListener('error', onError);
          proc.removeListener('exit', onExit);
          proc.stdout.removeAllListeners('data');
          proc.stderr.removeAllListeners('data');

          finished = true;

          if (error != null) {
            reject(error);
          } else {
            resolve(proc);
          }
        }

        proc.once('error', onError);

        proc.once('exit', onExit);

        yield sleep(1000);

        // do not poll if we are finished (error)
        if (finished) {
          return;
        }

        let retries = 0;

        do {
          const pids = yield (0, _portPid2.default)(port);

          if (pids.all.size === 0) {
            // nothing is listening (slow server) , or there is an error
          } else if (!pids.all.has(proc.pid)) {
            throw new PortIsOccupiedError([...pids.all][0], port);
          } else {
            return finish();
          }

          retries += 1;
          yield sleep(1000);
        } while (retries < retryLimit);

        throw new Error(`Your server is not listening on port ${port}`);
      } catch (e) {
        reject(e);
      }
    });

    return function (_x, _x2) {
      return _ref.apply(this, arguments);
    };
  })());
}