#!/usr/bin/env node
'use strict';

let start = (() => {
  var _ref = _asyncToGenerator(function* (dir, sourceDir) {
    const useBabili = !!parseInt(process.env.SPUST_USE_BABILI || '0', 10);
    const doNotOpenBrowser = process.env.SPUST_DO_NOT_OPEN_BROWSER ? true : false;
    const DEFAULT_PORT = parseInt(process.env.WEBPACK_PORT, 10) || 2999;
    const DEFAULT_SERVER_PORT = parseInt(process.env.PORT, 10) || 3000;
    const HOST = process.env.HOST || '0.0.0.0';
    const protocol = process.env.HTTPS === 'true' ? 'https' : 'http';
    let isCompiled = false;

    const webpackPort = yield (0, _WebpackDevServerUtils.choosePort)(HOST, DEFAULT_PORT);
    const serverPort = yield (0, _WebpackDevServerUtils.choosePort)(HOST, DEFAULT_SERVER_PORT);
    const urls = (0, _WebpackDevServerUtils.prepareUrls)(protocol, HOST, webpackPort);

    // set server port to env so we can use it in env variables for webpack
    process.env.PORT = String(serverPort);

    const serverManager = new _ServerProcessManager2.default(serverPort, dir);

    // close server manager on termination
    ['SIGINT', 'SIGTERM'].forEach(function (signal) {
      return process.on(signal, function () {
        serverManager.close();
        process.exit();
      });
    });

    let config = yield (0, _configure2.default)({
      workDir: dir,
      srcDir: sourceDir,
      env: 'development',
      serverManager,
      useBabili
    });

    // override configuration using own configure
    const configPath = (0, _path.resolve)(workDir, './spust.config.js');

    try {
      // dost spust.config.js file exist?
      yield (0, _fs.stat)(configPath);

      delete require.cache[require.resolve(configPath)];
      let customConfigure = require(configPath);

      config = customConfigure(config, {
        workDir: dir,
        srcDir: sourceDir,
        env: 'development',
        serverManager
      });
    } catch (e) {
      if (e.code !== 'ENOENT') {
        throw e;
      }
    }

    _rimraf2.default.sync((0, _path.resolve)(workDir, 'bundle'));

    const compiler = (0, _webpack2.default)([config.client, config.server]);
    const target = `http://localhost:${serverPort}`;

    const progress = new _Progress2.default(compiler.compilers);

    compiler.plugin('invalid', function () {
      // clear only after first done pass
      if (isInteractive && isCompiled) {
        (0, _clearConsole2.default)();
      }

      progress.start();
    });

    compiler.plugin('done', (() => {
      var _ref2 = _asyncToGenerator(function* (stats) {
        let hasWarnings = false;
        let warnings = '';

        progress.stop();

        if (isInteractive) {
          (0, _clearConsole2.default)();
        }

        const [clientStats, serverStats] = stats.stats;

        const clientMessages = (0, _formatWebpackMessages2.default)(clientStats.toJson({}, true));
        const serverMessages = (0, _formatWebpackMessages2.default)(serverStats.toJson({}, true));

        if (clientMessages.errors.length || serverMessages.errors.length) {
          console.log(_chalk2.default.red('⚠️  Failed to compile.\n'));
          console.log([...clientMessages.errors, ...serverMessages.errors].join('\n'));
          console.log(_chalk2.default.cyan('ℹ️  Keeping previous server instance running'));
          return;
        }

        if (clientMessages.warnings.length || serverMessages.warnings.length) {
          hasWarnings = true;
          warnings = [...clientMessages.warnings, ...serverMessages.warnings].join('\n');
        }

        const resultColor = hasWarnings ? 'yellow' : 'green';
        const resultState = isCompiled ? 'Updated' : 'Compiled';
        const resultMessage = hasWarnings ? 'with warnings' : 'successfully';

        console.log(_chalk2.default[resultColor](`🎉  ${resultState} ${resultMessage}!`));

        if (warnings.length > 1) {
          console.log(warnings);
        }

        // now check if in ServerManager has any errors from management
        // if yes, then output readable message about it
        if (!serverManager.isRunning()) {
          if (serverManager.lastSpawnErrors().length > 0) {
            console.log(_chalk2.default.red('⚠️  Server is not running, see errors:\n'));

            serverManager.lastSpawnErrors().forEach(function (e) {
              console.log(e.message);
              console.log();

              if (e instanceof _spawnServer.UnexpectedServerTerminationError) {
                console.log(e.stderr);
              } else if (e instanceof _spawnServer.ServerSpawnFailureError) {
                console.log(e.stderr);
              }

              console.log();
            });
          } else {
            console.log(_chalk2.default.red('⚠️  Server is not running because of an unkown error!\n'));
          }
        } else {
          if (serverManager.lastSpawnErrors().length > 0) {
            console.log(_chalk2.default.red('⚠️  Failed to spawn a new server for your backend, see errors:\n'));

            serverManager.lastSpawnErrors().forEach(function (e) {
              console.log(e.message);
              console.log();

              if (e instanceof _spawnServer.UnexpectedServerTerminationError) {
                console.log(e.stderr);
              } else if (e instanceof _spawnServer.ServerSpawnFailureError) {
                console.log(e.stderr);
              }

              console.log();
            });

            console.log(_chalk2.default.cyan('ℹ️  Keeping previous server running.'));
          }

          // if apps wasn't compiled yet, open browser (if user's did not tell otherwise)
          if (!doNotOpenBrowser && !isCompiled) {
            (0, _openBrowser2.default)(_url2.default.format({
              protocol,
              hostname: 'localhost',
              port: webpackPort
            }));
          }

          // print nice info :)
          (0, _printInstructions2.default)(urls);
        }

        // clear last spawn errors
        serverManager.clearSpawnErrors();

        // set flag that it has been successfully compiled
        // so in next builds it will result to Updated and not Compiled
        isCompiled = true;
      });

      return function (_x3) {
        return _ref2.apply(this, arguments);
      };
    })());

    progress.start();

    const devServer = new _webpackDevServer2.default(compiler, {
      contentBase: (0, _path.resolve)(__dirname, '../'),
      clientLogLevel: 'none',
      compress: true,
      historyApiFallback: {
        disableDotRule: true
      },
      hot: true,
      noInfo: true,
      overlay: false,
      proxy: [function () {
        return {
          context: '/static',
          logLevel: 'silent',
          target: _url2.default.format({
            protocol,
            hostname: 'localhost',
            port: webpackPort
          }),
          pathRewrite: function (path) {
            return path.replace('static', 'client/static');
          }
        };
      }, function () {
        return {
          context: function (pathname) {
            return pathname.indexOf('sockjs-node') === -1 && pathname.indexOf('hot-update.json') === -1 && pathname.indexOf('hot-update.js') === -1;
          },
          // wait for 10 minutes in case that server has error, so user can fix it :)
          proxyTimeout: (0, _ms2.default)('10 minutes'),
          target,
          logLevel: 'silent',
          onProxyReq: function (proxyReq) {
            if (proxyReq.getHeader('origin')) {
              proxyReq.setHeader('origin', target);
            }
          },
          onError: function (err, req, res) {
            const host = req.headers && req.headers.host;
            console.log(_chalk2.default.red('⚠️  Proxy error:') + ' Could not proxy request ' + _chalk2.default.cyan(req.url) + ' from ' + _chalk2.default.cyan(host) + ' to ' + _chalk2.default.cyan(target) + '.');
            console.log('See https://nodejs.org/api/errors.html#errors_common_system_errors for more information (' + _chalk2.default.cyan(err.code) + ').');
            console.log();

            // And immediately send the proper error response to the client.
            // Otherwise, the request will eventually timeout with ERR_EMPTY_RESPONSE on the client side.
            if (res.writeHead && !res.headersSent) {
              res.writeHead(500);
            }
            res.end('⚠️  Proxy error: Could not proxy request');
          },
          secure: false,
          changeOrigin: true,
          ws: true,
          xfwd: true
        };
      }],
      publicPath: config.client.output.publicPath,
      quiet: true,
      setup(app) {
        // This lets us open files from the runtime error overlay.
        app.use((0, _middleware2.default)());
      },
      watchOptions: {
        ignored: /bundle|node_modules/
      }
    });

    devServer.listen(webpackPort, HOST, function (err) {
      if (err) {
        return console.log(err);
      }

      if (isInteractive) {
        (0, _clearConsole2.default)();
      }

      console.log(_chalk2.default.cyan('🚀  Starting the development server...'));

      if (!doNotOpenBrowser) {
        console.log(_chalk2.default.cyan('ℹ️  I will open the browser after successful compilation.'));
      }

      console.log();
    });

    ['SIGINT', 'SIGTERM'].forEach(function (sig) {
      process.on(sig, function () {
        devServer.close();
        process.exit();
      });
    });
  });

  return function start(_x, _x2) {
    return _ref.apply(this, arguments);
  };
})();

var _path = require('path');

var _fs = require('mz/fs');

var _WebpackDevServerUtils = require('react-dev-utils/WebpackDevServerUtils');

var _chalk = require('chalk');

var _chalk2 = _interopRequireDefault(_chalk);

var _clearConsole = require('react-dev-utils/clearConsole');

var _clearConsole2 = _interopRequireDefault(_clearConsole);

var _middleware = require('react-error-overlay/middleware');

var _middleware2 = _interopRequireDefault(_middleware);

var _formatWebpackMessages = require('react-dev-utils/formatWebpackMessages');

var _formatWebpackMessages2 = _interopRequireDefault(_formatWebpackMessages);

var _ms = require('ms');

var _ms2 = _interopRequireDefault(_ms);

var _openBrowser = require('react-dev-utils/openBrowser');

var _openBrowser2 = _interopRequireDefault(_openBrowser);

var _rimraf = require('rimraf');

var _rimraf2 = _interopRequireDefault(_rimraf);

var _url = require('url');

var _url2 = _interopRequireDefault(_url);

var _webpack = require('webpack');

var _webpack2 = _interopRequireDefault(_webpack);

var _webpackDevServer = require('webpack-dev-server');

var _webpackDevServer2 = _interopRequireDefault(_webpackDevServer);

var _configure = require('./configure');

var _configure2 = _interopRequireDefault(_configure);

var _printInstructions = require('./printInstructions');

var _printInstructions2 = _interopRequireDefault(_printInstructions);

var _Progress = require('./Progress');

var _Progress2 = _interopRequireDefault(_Progress);

var _ServerProcessManager = require('./ServerProcessManager');

var _ServerProcessManager2 = _interopRequireDefault(_ServerProcessManager);

var _spawnServer = require('./spawnServer');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

/* eslint-disable import/first */

process.env.BABEL_ENV = 'development';
process.env.NODE_ENV = 'development';

// crash on unhandledRejection
process.on('unhandledRejection', err => {
  throw err;
});

const isInteractive = process.stdout.isTTY;
const srcDir = process.argv.slice(3).find(arg => !arg.startsWith('--')) || 'src';
const workDir = process.argv[2];

if (workDir == null) {
  throw new Error('Please provide workDir path as the first argument');
}

start(workDir, srcDir).catch(e => {
  console.log(e);
  process.exit(1);
});